#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <libgen.h>
#include <dirent.h>
#include <pwd.h>
#include <grp.h>
#include <time.h>
//----------------------------------------------------------------------
#define BUFFER_SIZE     131072
#define LIST_SIZE       8000
//----------------------------------------------------------------------
static void     usage (void);
static int      compare (const void *p1, const void *p2);
//----------------------------------------------------------------------
//
// Static variables to avoid stack overflow.
//
static char     buffer[BUFFER_SIZE];
static char    *list[LIST_SIZE];
//----------------------------------------------------------------------
int
main (int argc, char *argv[], char *envp[])
{
  int             option_a = 0;
  int             option_l = 0;
  int             opt;
//  extern char   *optarg;              // not used.
  extern int      optind;
  extern int      optopt;
  struct stat     file_status;
  DIR            *dp;
  struct dirent  *dir;
  int             b;		// Buffer index.
  int             l;		// List index.
  int             len;		// Name length.
  char           *path = NULL;
  char            pathname[PATH_MAX];
  struct passwd  *pws;
  struct group   *grs;
  struct tm      *tms;
  //
  // Check for options.
  //
  while ((opt = getopt (argc, argv, ":al")) != -1)
    {
      switch (opt)
	{
	case 'l':
	  option_l = 1;
	  break;
	case 'a':
	  option_a = 1;
	  break;
	case '?':
	  fprintf (stderr, "Unknown option -%c.\n", optopt);
	  usage ();
	  return (1);
	  break;
	case ':':
	  fprintf (stderr, "Missing argument for option -%c\n", optopt);
	  usage ();
	  return (1);
	  break;
	default:
	  fprintf (stderr, "Getopt problem: unknown option %c\n", opt);
	  return (1);
	}
    }
  //
  // If no arguments are present, at least the current directory is
  // read.
  //
  if (optind == argc)
    {
      //
      // There are no more arguments. Replace the program name,
      // corresponding to `argv[0]', with the current directory
      // path string.
      //
      argv[0] = ".";
      argc = 1;
      optind = 0;
    }
  //
  // This is a very simplified `ls': if there is only a name
  // and it is a directory, the directory content is taken as
  // the new `argv[]' array.
  //
  if (optind == (argc - 1))
    {
      //
      // There is a request for a single name. Test if it exists
      // and if it is a directory.
      //
      if (stat (argv[optind], &file_status) != 0)
	{
	  fprintf (stderr, "File \"%s\" does not exist!\n",
		   argv[optind]);
	  return (2);
	}
      //
      if (S_ISDIR (file_status.st_mode))
	{
	  //
	  // Save the directory inside the `path' pointer.
	  //
	  path = argv[optind];
	  //
	  // Open the directory.
	  //
	  dp = opendir (argv[optind]);
	  if (dp == NULL)
	    {
	      perror (argv[optind]);
	      return (3);
	    }
	  //
	  // Read the directory and fill the buffer with names.
	  //
	  b = 0;
	  l = 0;
	  while ((dir = readdir (dp)) != NULL)
	    {
	      len = strlen (dir->d_name);
	      //
	      // Check if the buffer can hold it.
	      //
	      if ((b + len + 1) > BUFFER_SIZE)
		{
		  fprintf (stderr, "not enough memory\n");
		  break;
		}
	      //
	      // Consider the directory item only if there is
	      // a valid name. If it is empty, just ignore it.
	      //
	      if (len > 0)
		{
		  strcpy (&buffer[b], dir->d_name);
		  list[l] = &buffer[b];
		  b += len + 1;
		  l++;
		}
	    }
	  //
	  // Close the directory.
	  //
	  closedir (dp);
	  //
	  // Sort the list.
	  //
	  qsort (list, (size_t) l, sizeof (char *), compare);
	  //
	  // Convert the directory list into a new `argv[]' array,
	  // with a valid `argc'. The variable `optind' must be
	  // reset to the first element index, because there is
	  // no program name inside the new `argv[]' at index zero.
	  //
	  argv = list;
	  argc = l;
	  optind = 0;
	}
    }
  //
  // Scan arguments, or list converted into `argv[]'.
  //
  for (; optind < argc; optind++)
    {
      if (argv[optind][0] == '.')
	{
	  //
	  // Current name starts with `.'.
	  //
	  if (!option_a)
	    {
	      //
	      // Do not show name starting with `.'.
	      //
	      continue;
	    }
	}
      //
      // Build the pathname.
      //
      if (path == NULL)
	{
	  strcpy (&pathname[0], argv[optind]);
	}
      else
	{
	  strcpy (pathname, path);
	  strcat (pathname, "/");
	  strcat (pathname, argv[optind]);
	}
      //
      // Check if file exists, reading status.
      //
      if (stat (pathname, &file_status) != 0)
	{
	  fprintf (stderr, "File \"%s\" does not exist!\n", pathname);
	  return (2);
	}
      //
      // Show file name.
      //
      if (option_l)
	{
	  //
	  // Print the file type.
	  //
	  if (S_ISBLK (file_status.st_mode))
	    printf ("b");
	  else if (S_ISCHR (file_status.st_mode))
	    printf ("c");
	  else if (S_ISFIFO (file_status.st_mode))
	    printf ("p");
	  else if (S_ISREG (file_status.st_mode))
	    printf ("-");
	  else if (S_ISDIR (file_status.st_mode))
	    printf ("d");
	  else if (S_ISLNK (file_status.st_mode))
	    printf ("l");
	  else if (S_ISSOCK (file_status.st_mode))
	    printf ("s");
	  else
	    printf ("?");
	  //
	  // Print permissions.
	  //
	  if (S_IRUSR & file_status.st_mode)
	    printf ("r");
	  else
	    printf ("-");
	  if (S_IWUSR & file_status.st_mode)
	    printf ("w");
	  else
	    printf ("-");
	  if (S_IXUSR & file_status.st_mode)
	    printf ("x");
	  else
	    printf ("-");
	  if (S_IRGRP & file_status.st_mode)
	    printf ("r");
	  else
	    printf ("-");
	  if (S_IWGRP & file_status.st_mode)
	    printf ("w");
	  else
	    printf ("-");
	  if (S_IXGRP & file_status.st_mode)
	    printf ("x");
	  else
	    printf ("-");
	  if (S_IROTH & file_status.st_mode)
	    printf ("r");
	  else
	    printf ("-");
	  if (S_IWOTH & file_status.st_mode)
	    printf ("w");
	  else
	    printf ("-");
	  if (S_IXOTH & file_status.st_mode)
	    printf ("x");
	  else
	    printf ("-");
	  //
	  // Print links.
	  //
	  printf (" %3i", (int) file_status.st_nlink);
	  //
	  // Print owner.
	  //
	  pws = getpwuid (file_status.st_uid);
	  if (pws == NULL)
	    {
	      printf (" %i", (int) file_status.st_uid);
	    }
	  else
	    {
	      printf (" %s", pws->pw_name);
	    }
	  //
	  // Print group.
	  //
	  grs = getgrgid (file_status.st_gid);
	  if (grs == NULL)
	    {
	      printf (" %i", (int) file_status.st_gid);
	    }
	  else
	    {
	      printf (" %s", grs->gr_name);
	    }
	  //
	  // Print file size or device major-minor.
	  //
	  if (S_ISBLK (file_status.st_mode)
	      || S_ISCHR (file_status.st_mode))
	    {
	      printf (" %3i,", (int) major (file_status.st_rdev));
	      printf (" %3i", (int) minor (file_status.st_rdev));
	    }
	  else
	    {
	      printf (" %8i", (int) file_status.st_size);
	    }
	  //
	  // Print modification date and time.
	  //
	  tms = localtime (&(file_status.st_mtime));
	  printf (" %4u-%02u-%02u %02u:%02u",
		  tms->tm_year, tms->tm_mon, tms->tm_mday,
		  tms->tm_hour, tms->tm_min);
	  //
	  // Print file name, but with no additional path.
	  //
	  printf (" %s\n", argv[optind]);
	}
      else
	{
	  //
	  // Just show the file name and go to the next line.
	  //
	  printf ("%s\n", argv[optind]);
	}
    }
  //
  // All done.
  //
  return (0);
}

//----------------------------------------------------------------------
static void
usage (void)
{
  fprintf (stderr, "Usage: ls [OPTION] [FILE]...\n");
}

//----------------------------------------------------------------------
static int
compare (const void *p1, const void *p2)
{
  return (strcmp (*((char **) p1), *((char **) p2)));
}
