#include <kernel/fs.h>
#include <errno.h>
#include <kernel/lib_k.h>
#include <kernel/dev.h>
//----------------------------------------------------------------------
inode_t *
inode_get (dev_t device, ino_t ino)
{
    sb_t             *sb;
    inode_t          *inode;
    unsigned long int start;
    size_t            size;
    ssize_t           n;
    int               status;
    //
    // Verify if the root file system inode was requested.
    //
    if (device == 0 && ino == 1)
      {
        //
        // Get root file system inode.
        //
        inode = inode_reference (device, ino);
        if (inode == NULL)
          {
            //
            // The file system root directory inode is not yet loaded:
            // get the first super block.
            //
            sb = sb_reference ((dev_t) 0);
            if (sb == NULL || sb->device == 0)
              {
                //
                // This error should never happen.
                //
                errset (EUNKNOWN);      // Unknown error.
                return (NULL);
              }
            //
            // Load the file system root directory inode (recursive
            // call).
            //
            inode = inode_get (sb->device, (ino_t) 1);
            if (inode == NULL)
              {
                //
                // This error should never happen.
                //
                errset (EUNKNOWN);      // Unknown error.
                return (NULL);
              }
            //
            // Return the directory inode.
            //
            return (inode);
          }
        else
          {
            //
            // The file system root directory inode is already
            // available.
            //
            if (inode->references >= INODE_MAX_REFERENCES)
              {
                errset (ENFILE); // Too many files open in system.
                return (NULL);
              }
            else
              {
                inode->references++;
                return (inode);
              }
          }
      }
    //
    // A common device-inode pair was requested: try to find an already
    // cached inode.
    //
    inode = inode_reference (device, ino);
    if (inode != NULL)
      {
        if (inode->references >= INODE_MAX_REFERENCES)
          {
            errset (ENFILE);    // Too many files open in system.
            return (NULL);
          }
        else
          {
            inode->references++;
            return (inode);
          }
      }
    //
    // The inode is not yet available: get super block.
    //
    sb = sb_reference (device);
    if (sb == NULL)
      {
        errset (ENODEV);        // No such device.
        return (NULL);
      }
    //
    // The super block is available, but the inode is not yet cached.
    // Verify if the inode map reports it as allocated.
    //
    status = sb_inode_status (sb, ino);
    if (!status)
      {
        //
        // The inode is not allocated and cannot be loaded.
        //
        errset (ENOENT);        // No such file or directory.
        return (NULL);
      }
    //
    // The inode was not already cached, but is considered as allocated
    // inside the inode map. Find a free slot to load the inode inside
    // the inode table (in memory).
    //
    inode = inode_reference ((dev_t) -1, (ino_t) -1);
    if (inode == NULL)
      {
        errset (ENFILE);        // Too many files open in system.
        return (NULL);
      }
    //
    // A free inode slot was found. The inode must be loaded.
    // Calculate the memory inode size, to be saved inside the file
    // system: the administrative inode data, as it is saved inside
    // the file system. The `inode_t' type is bigger than the real
    // inode administrative size, because it contains more data, that is
    // not saved on disk.
    //
    size   = offsetof (inode_t, sb);
    //
    // Calculating start position for read.
    //
    // [1] Boot block.
    // [2] Super block.
    // [3] Inode bit map.
    // [4] Zone bit map.
    // [5] Previous inodes: consider that the inode zero is
    //     present in the inode map, but not in the inode
    //     table.
    //
    start  = 1024;                           // [1]
    start += 1024;                           // [2]
    start += (sb->map_inode_blocks * 1024);  // [3]
    start += (sb->map_zone_blocks * 1024);   // [4]
    start += ((ino -1) * size);              // [5]
    //
    // Read inode from disk.
    //
    n = dev_io ((pid_t) -1, device, DEV_READ, start, inode, size, NULL);
    if (n != size)
      {
        errset (EIO);                   // I/O error.
        return (NULL);
      }
    //
    // The inode was read: add some data to the working copy in memory.
    //
    inode->sb            = sb;
    inode->sb_attached   = NULL;
    inode->ino           = ino;
    inode->references    = 1;
    inode->changed       = 0;
    //
    inode->blkcnt        = inode->size;
    inode->blkcnt       /= sb->blksize;
    if (inode->size % sb->blksize)
      {
        inode->blkcnt++;
      }
    //
    inode->pipe_dir       = 1;   // Pipes must start with write.
    inode->pipe_off_read  = 0;
    inode->pipe_off_write = 0;
    inode->pipe_ref_read  = 0;
    inode->pipe_ref_write = 0;
    //
    // Return the inode pointer.
    //
    return (inode);
}
