#include <kernel/proc.h>
#include <kernel/net.h>
#include <kernel/net/route.h>
#include <errno.h>
#include <arpa/inet.h>
#include <sys/os32.h>
//----------------------------------------------------------------------
#define DEBUG 0
//----------------------------------------------------------------------
ssize_t
s_recv (pid_t pid, int sfdn, void *buffer, size_t length, int flags)
{
  fd_t     *sfd;
  h_addr_t  local;
  int       i;          // IP table index.
  size_t    size_read;
  //
  // Get file descriptor and verify that it is a socket.
  //
  sfd = fd_reference (pid, &sfdn);
  if (sfd == NULL || sfd->file == NULL)
    {
      errset (EBADF);		// Bad file descriptor.
      return ((ssize_t) -1);
    }
  if (sfd->file->sock == NULL)
    {
      errset (ENOTSOCK);	// Not a socket.
      return ((ssize_t) -1);
    }
  //
  // Verify to have a valid buffer pointer.
  //
  if (buffer == NULL)
    {
      errset (EINVAL);
      return ((ssize_t) -1);
    }
  //
  //
  //
  if (sfd->file->sock->family == AF_INET)                    // INET
    {
      //
      // AF_INET requires at least the destination address.
      //
      if (sfd->file->sock->addr == 0)
        {
          errset (EDESTADDRREQ);
          return ((ssize_t) -1);
        }
      //
      // Get the local address, from the routing table, based
      // on the remote address.
      //
      local = route_remote_to_local (sfd->file->sock->addr);
      if (local == ((h_addr_t) -1))
        {
          errset (errno);
          return ((ssize_t) -1);
        }
      //
      if (sfd->file->sock->type == SOCK_RAW)                 // RAW
        {
          if (sfd->file->sock->protocol == IPPROTO_ICMP)     // ICMP
            {
              //
              // Scan the ip_table[] to find an ICMP packet
              // that was not already seen by the socket.
              //
              for (i = 0; i < IP_MAX_PACKETS; i++)
                {
                  if (ip_table[i].packet.header.src
                       == htonl (sfd->file->sock->addr)
                      && ip_table[i].packet.header.dst == htonl (local)
                      && ip_table[i].packet.header.protocol
                          == IP_PROTOCOL_ICMP
                      && ip_table[i].clock
                          != sfd->file->sock->read.clock[i])
                    {
                      //
                      // This ICMP RAW packet is new for the
                      // socket: save the clock time, so that the
                      // same packet is not read again.
                      //
                      sfd->file->sock->read.clock[i]
                       = ip_table[i].clock;
                      //
                      // Copy the packet.
                      //
                      size_read = min (ip_table[i].packet.header.length,
                                       length);
                      //
                      memcpy (buffer, ip_table[i].packet.octet,
                              size_read);
                      //
                      return ((ssize_t) size_read);
                    }
                }
            }
          else
            {
              //
              // Unsupported protocol.
              //
              errset (EPROTONOSUPPORT);
              return ((ssize_t) -1);
            }
        }
      else
        {
          //
          // Unsupported type.
          //
          errset (EPROTONOSUPPORT);
          return ((ssize_t) -1);
        }
    }
  else
    {
      //
      // Unsupported family.
      //
      errset (EAFNOSUPPORT);
      return ((ssize_t) -1);
    }
  //
  // If we are here, there are no more packets to read, so the
  // process should go to sleep.
  //
  proc_table[pid].status = PROC_SLEEPING;
  proc_table[pid].ret = 0;
  proc_table[pid].wakeup_events = WAKEUP_EVENT_SOCK_READ;
  if (DEBUG)
    {
      k_printf ("[%s] PID %i goes to sleep "
                "waiting to receive for a socket.\n",
                __FILE__, pid);
    }
  //
  // Nothing was received.
  //
  return ((ssize_t) 0);
}
