#include <kernel/proc.h>
#include <kernel/lib_s.h>
#include <errno.h>
#include <fcntl.h>
//----------------------------------------------------------------------
#define DEBUG 0
//----------------------------------------------------------------------
ssize_t
s_write (pid_t pid, int fdn, void *buffer, size_t count)
{
  proc_t         *ps;
  fd_t           *fd;
  ssize_t         size_written;
  int             status;
  //
  // Get process.
  //
  ps = proc_reference (pid);
  //
  // Get file descriptor.
  //
  fd = fd_reference (pid, &fdn);
  if (fd == NULL || fd->file == NULL
      || (fd->file->inode == NULL && fd->file->sock == NULL))
    {
      //
      // The file descriptor pointer is not valid.
      //
      errset (EBADF);		// Bad file descriptor.
      return ((ssize_t) - 1);
    }
  //
  // Check if it is opened for write.
  //
  if (!(fd->file->oflags & O_WRONLY))
    {
      //
      // The file is not opened for write.
      //
      errset (EINVAL);		// Invalid argument.
      return ((ssize_t) - 1);
    }
  //
  // Check if it is a directory inode: a directory can be
  // read as a file descriptor, but cannot be written.
  //
  if (fd->file->inode != NULL && (fd->file->inode->mode & S_IFDIR))
    {
      errset (EISDIR);		// Is a directory.
      return ((ssize_t) - 1);
    }
  //
  // It should be a valid type of file or socket to be written.
  // Check if it is a file opened in append mode: if so, must move
  // the write offset to the end.
  //
  if (fd->file->inode != NULL && (fd->fl_flags & O_APPEND))
    {
      fd->file->offset = fd->file->inode->size;
    }
  //
  // Check the kind of socket/file to be written and write it.
  //
  if (fd->file->sock != NULL)
    {
      //
      // Send it.
      //
      size_written = s_send (pid, fdn, buffer, count, 0);
    }
  else if (fd->file->inode->mode & S_IFBLK ||
           fd->file->inode->mode & S_IFCHR)
    {
      //
      // A device is to be written.
      //
      size_written = dev_io (pid, (dev_t) fd->file->inode->direct[0],
			     DEV_WRITE, (off_t) fd->file->offset,
			     buffer, count, NULL);
    }
  else if (fd->file->inode->mode & S_IFREG)
    {
      //
      // A regular file is to be written.
      //
      size_written = inode_file_write (fd->file->inode,
				       fd->file->offset, buffer, count);
    }
  else if (fd->file->inode->mode & S_IFIFO)
    {
      //
      // A pipe is to be written.
      //
      size_written = inode_pipe_write (fd->file->inode, buffer, count);
      if (size_written == 0)
	{
	  if (fd->file->inode->pipe_ref_read == 0)
	    {
	      //
	      // No read will be done anymore. Tell to
	      // the process.
	      //
	      status = s_kill ((pid_t) 0, pid, SIGPIPE);
	      if (status < 0)
		{
		  errset (EPIPE);
		  return ((ssize_t) - 1);
		}
	      else
		{
		  //
		  // Wake up processes waiting to read.
		  //
		  proc_wakeup_pipe_read (fd->file->inode);
		}
	    }
	  else
	    {
	      //
	      // Go to sleep.
	      //
	      proc_table[pid].status = PROC_SLEEPING;
	      proc_table[pid].ret = 0;
	      proc_table[pid].wakeup_inode = fd->file->inode;
	      proc_table[pid].wakeup_events = WAKEUP_EVENT_PIPE_WRITE;
	      if (DEBUG)
		{
		  k_printf ("[%s] PID %i goes to sleep "
			    "waiting to write inside a pipe.\n",
			    __FILE__, pid);
		}
	    }
	}
      else
	{
	  //
	  // Wake up processes waiting to read.
	  //
	  proc_wakeup_pipe_read (fd->file->inode);
	}
    }
  else
    {
      //
      // Unsupported file type.
      //
      errset (E_FILE_TYPE_UNSUPPORTED);	//File type unsupported.
      return ((ssize_t) - 1);
    }
  //
  // Update the file descriptor internal offset, but only if it
  // is a file.
  //
  if (fd->file->sock == NULL && size_written > 0)
    {
      fd->file->offset += size_written;
    }
  //
  // Just return the size written, even if it is an error.
  //
  return (size_written);
}
