#ifndef _KERNEL_NET_H
#define _KERNEL_NET_H    1
//----------------------------------------------------------------------
#include <stdint.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <lib/sys/os32.h>
//----------------------------------------------------------------------
#define NET_PROT_IP     0x0800
#define NET_PROT_ARP    0x0806
#define NET_PROT_RARP   0x8035
//
#define NET_ETHERNET_ADDRESS_LENGTH   6
#define NET_IP_ADDRESS_LENGTH         4
//
#define NET_DEV_NULL               0x0000
#define NET_DEV_LOOP               0x0100
#define NET_DEV_LOOPBACK           0x0101
#define NET_DEV_ETH                0x0200
#define NET_DEV_ETH_NE2K           0x0201
//
#define NET_MAX_DEVICES               4
#define NET_MAX_BUFFERS              64
//
#define NET_ETHERNET_MIN_PACKET_SIZE  46
#define NET_ETHERNET_MAX_PACKET_SIZE  1500
#define NET_ETHERNET_HEADER_SIZE      14
#define NET_ETHERNET_MAX_FRAME_SIZE   (NET_ETHERNET_MAX_PACKET_SIZE \
                                       +NET_ETHERNET_HEADER_SIZE)
//
#define NET_IP_MIN_HEADER_SIZE  20
#define NET_IP_MIN_PACKET_SIZE  576
#define NET_IP_MAX_PACKET_SIZE  NET_ETHERNET_MAX_PACKET_SIZE      // [1]
#define NET_IP_MAX_DATA_SIZE    (NET_IP_MAX_PACKET_SIZE \
                                 -NET_IP_MIN_HEADER_SIZE)
//
// [1] The IP max packet size should be 65535, but as os32 does not
//     accept fragmented packets, the maximum value depends on the
//     Ethernet packet size. But this limitation might change, so
//     there is also the value NET_MTU, that depends from the
//     real physical network limitations.
//
#define NET_MTU                 NET_ETHERNET_MAX_PACKET_SIZE
//----------------------------------------------------------------------
//
// IP header.
//
typedef struct {
  uint8_t   ihl      : 4,
            version  : 4;
  uint8_t   service;
  uint16_t  length;
  uint16_t  id;
  uint16_t  fragment;    // [flg][fragment_offset] 3+13 bit
  uint8_t   ttl;
  uint8_t   protocol;
  uint16_t  checksum;
  in_addr_t src;         // Network byte order.
  in_addr_t dst;         // Network byte order.
} __attribute__ ((packed)) net_ip_header_t;
//
// IP unfragmented packet, but it is not known if there are IP options,
// so it is not possible to define the space after the minimal header.
//
typedef union {
  uint8_t             octet[NET_IP_MAX_PACKET_SIZE];
  net_ip_header_t     header;
} __attribute__ ((packed)) net_ip_packet_t;
//
// Ethernet header.
//
typedef struct {
  uint8_t  dst[6];
  uint8_t  src[6];
  uint16_t type;
} __attribute__ ((packed)) net_ethernet_header_t;
//
// Ethernet frame.
//
// .-----------------------------------------------------------------.
// | .                                                               |
// |-----------------------------------------------------------------|
// | .octect[]                                                       |
// |----------.------------------------------------------------------|
// | .header  | .packet                                              |
// |----------+------------------------------------------------------|
// | ...      | .packet.octect[]                                     |
// |----------+----------------.-------------------------------------|
// |          | .packet.header | / / / / / / / / / / / / / / / / / / |
// `----------'----------------'-------------------------------------'
//
typedef union {
  //
  uint8_t octet[NET_ETHERNET_MAX_FRAME_SIZE];
  //
  struct {
    net_ethernet_header_t header;
    union {
      //
      uint8_t             octet[NET_IP_MAX_PACKET_SIZE];
      //
      net_ip_header_t     header;
      //
    } __attribute__ ((packed)) packet;
  } __attribute__ ((packed));
  //
} __attribute__ ((packed)) net_ethernet_frame_t;
//----------------------------------------------------------------------
//
// Ethernet buffer.
//
typedef struct {
  clock_t              clock;
  size_t               size;
  net_ethernet_frame_t frame;
} __attribute__ ((packed)) net_ethernet_buffer_t;
//
// Loopback buffer.
//
typedef struct {
  clock_t  clock;
  size_t   size;
  union {
    uint8_t          octet[NET_MTU];
    net_ip_header_t  header;
  } packet;
} __attribute__ ((packed)) net_loopback_buffer_t;       // [2]
//
// [2] The structure net_ip_packet_t is not used here, because
//     it referes to unfragmented packets, where here the max
//     size might be less.
//
//----------------------------------------------------------------------
//
// Network interfaces table structure
//
typedef struct {
  unsigned int type;
  h_addr_t     ip;      // IPv4 address in host byte order.
  uint8_t      m;       // Short netmask.
  union {
    //
    // Ethernet type data:
    //
    struct {
      uint8_t               mac[6];
      uintptr_t             base_io;
      unsigned char         irq;
      net_ethernet_buffer_t buffer[NET_MAX_BUFFERS];
    } ethernet;
    //
    // Loopback type data:
    //
    struct {
      net_loopback_buffer_t buffer[NET_MAX_BUFFERS];
    } loopback;
  };
} net_t;
//
// [2] The structure net_ip_packet_t is not used here, becase
//     it refere to unfragmented packets, where here the max
//     size might be less.
//
extern net_t net_table[NET_MAX_DEVICES];
//----------------------------------------------------------------------
int  net_rx              (void);
void net_init            (void);
int  net_index           (h_addr_t ip);
int  net_ethernet_index  (uintptr_t io, uint8_t mac[6], h_addr_t ip);
int  net_ip_tx           (h_addr_t src, h_addr_t dst,
                          const void *packet, size_t size);

net_ethernet_buffer_t *net_ethernet_buffer (int n);
net_loopback_buffer_t *net_loopback_buffer (int n);

uint32_t net_ip_h       (uint8_t a, uint8_t b, uint8_t c, uint8_t d);
uint32_t net_ip_n       (uint8_t a, uint8_t b, uint8_t c, uint8_t d);
uint32_t net_mask_h     (int m);
uint32_t net_mask_n     (int m);
void     net_print      (void);
//----------------------------------------------------------------------
#endif

