#include <kernel/net.h>
#include <kernel/net/ethernet.h>
#include <kernel/net/arp.h>
#include <kernel/net/icmp.h>
#include <kernel/net/ip.h>
#include <sys/os32.h>
#include <kernel/lib_k.h>
#include <errno.h>
#include <arpa/inet.h>
//----------------------------------------------------------------------
#define DEBUG 0
//----------------------------------------------------------------------
int
ip_rx (int n, int b)
{
  net_ip_header_t *header;
  uint8_t         *packet;
  uint16_t         checksum;
  size_t           size_header;
  ip_t            *ip_table_item;
  int              i;
  //
  //
  //
  if (n >= NET_MAX_DEVICES || n < 0)
    {
      errset (EINVAL);          // Invalid argument.
      return (-1);
    }
  if (b >= NET_MAX_BUFFERS || b < 0)
    {
      errset (EINVAL);          // Invalid argument.
      return (-1);
    }
  //
  // Get the packet link.
  //
  if (net_table[n].type & NET_DEV_LOOP)
    {
      packet = (uint8_t *) &net_table[n].loopback.buffer[b].packet;
    }
  else if (net_table[n].type & NET_DEV_ETH)
    {
      //
      // It is Ethernet, but must also have an IP packet inside!
      //
      if (ntohs (net_table[n].ethernet.buffer[b].frame.header.type)
          != NET_PROT_IP)
        {
          errset (EINVAL);          // Invalid argument.
          return (-1);
        }
      packet = (uint8_t *)
               &net_table[n].ethernet.buffer[b].frame.packet;
    }
  else
    {
      errset (EINVAL);          // Invalid argument.
      return (-1);
    }
  //
  // The beguinning of the packet contains the IP header.
  //
  header = (void *) packet;
  //
  // Verify IP header checksum: it is also calculated the real header
  // size.
  //
  size_header = header->ihl * 4;
  checksum    = ip_checksum ((uint16_t *) header, size_header);
  if (checksum == 0xFFFF || checksum == 0x0000)
    {
      ; // k_printf ("checksum ok\n");
    }
  else
    {
      k_printf ("BAD CHECKSUM: %04x\n", checksum);
      return (0);
    }
  //
  // Is it a fragment? As we are not able to manage fragments,
  // we just check that it is all zero, ignoring the bit `DF'.
  // That is why we use a mask 0xBFFF.
  //
  if ((ntohs (header->fragment) & 0xBFFF) != 0)
    {
      //
      // Sorry, we don't manage fragments.
      //
      k_printf ("Sorry: we don't manage IP fragments: %04x\n", (ntohs (header->fragment)));
      return (0);
    }
  else
    {
      //
      // Find a place inside the IP table.
      //
      ip_table_item = ip_reference ();
      //
      // Copy the packet inside the ip_table[] item, then update
      // the link to the data start inside the packet and the
      // clock_t timestamp.
      //
      memcpy (ip_table_item->packet.octet, packet,
              ntohs (header->length));
      ip_table_item->pdu4  = ip_table_item->packet.octet;
      ip_table_item->pdu4 += (header->ihl * 4);
      ip_table_item->clock = k_clock ();
    }
  //
  // Now do something with the data inside the `ip_table[]'.
  //
  for (i = 0; i < IP_MAX_PACKETS; i++)
    {
      if (ip_table[i].clock != 0)
        {
          //
          if (ip_table[i].kernel_serviced != ip_table[i].clock
              && ip_table[i].packet.header.protocol == IP_PROTOCOL_ICMP)
            {
              icmp_rx (i);
            }
          else
            {
              //
              // At the moment, no other IP protocol managed internally.
              //
              ip_table[i].kernel_serviced = ip_table[i].clock;
            }
        }
    }
  //
  //
  //
  return (0);
}
